from . import panel_llm_monitor
from . import panel_trace_monitor
from . import panel_autoboard
from . import panel_seedboard
from . import panel_observability
from . import generator_templates

# This is just an example, uncomment to enable
# from . import panel_example_custom_board
